

from sqlalchemy import Column, Integer, String, Float

from .base import BaseModel


class WOFRegion(BaseModel):

    __tablename__ = 'wof_region'

    wof_id = Column(Integer, primary_key=True)

    wof_continent_id = Column(Integer)

    wof_country_id = Column(Integer)

    fips_code = Column(String)

    gn_id = Column(Integer)

    gp_id = Column(Integer)

    hasc_id = Column(String)

    iso_id = Column(String)

    unlc_id = Column(String)

    wd_id = Column(String)

    name = Column(String)

    name_abbr = Column(String)

    country_iso = Column(String)

    name_a0 = Column(String)

    latitude = Column(Float)

    longitude = Column(Float)

    population = Column(Integer)

    area_m2 = Column(Float)

    def __repr__(self):
        return '%s<%s, %s, wof:%d>' % (
            self.__class__.__name__,
            self.name, self.name_a0, self.wof_id,
        )





import os
import yaml


def read_yaml(from_path, file_name):
    """Open a YAML file relative to the passed path.

    Args:
        from_path (str)
        file_name (str)

    Returns: dict
    """
    path = os.path.join(os.path.dirname(from_path), file_name)

    with open(path, 'r') as fh:
        return yaml.load(fh)





import pytest

from litecoder.models import WOFRegion

from tests.utils import read_yaml


def yield_cases():
    """Generate cases from YAML file.
    """
    cases = read_yaml(__file__, 'test_us_state_index.yml')

    for group in cases:

        queries = group['query']

        if type(queries) is str:
            queries = [queries]

        for query in queries:
            yield query, group['matches']


@pytest.mark.parametrize('query,matches', yield_cases())
def test_cases(state_idx, query, matches):

    res = state_idx[query]

    ids = [r["wof_id"] for r in res]

    assert sorted(ids) == sorted(matches)


states = WOFRegion.query.filter(WOFRegion.country_iso=='US')


@pytest.mark.parametrize('state', states)
def test_all(state_idx, state):
    """Smoke test N most populous cities.
    """
    res = state_idx[state.name]
    res_ids = [r["wof_id"] for r in res]

    assert state.wof_id in res_ids




from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.inspection import inspect

from ..db import session


class BaseModel:

    @classmethod
    def column_names(cls):
        """Get a list of column names.
        """
        return cls.__table__.columns.keys()

    def __iter__(self):
        """Generate column / value tuples.

        Yields: (key, val)
        """
        md = inspect(self.__class__)

        for key in md.attrs.keys():
            yield (key, getattr(self, key))

        for key in md.relationships.keys():

            # Try to hydrate FK.
            row = getattr(self, key)

            # If exists, make nested dict.
            if row:
                yield (key, dict(getattr(self, key)))


BaseModel = declarative_base(cls=BaseModel)

BaseModel.query = session.query_property()





import pytest

from litecoder.usa import USCityIndex


# TODO: Possible to speed this up?
def test_bare_name_bloclist():
    """Blocklisted bare names should be omitted from index.
    """
    idx = USCityIndex(bare_name_blocklist=['Washington', 'New York'])
    idx.build()

    blocked = (
        'Washington',
        'washington',
        'Washington, USA',
        'New York',
        'new york',
        'New York, USA'
    )

    for query in blocked:
        assert not idx[query]

    assert idx['Washington DC'][0].data.wof_id == 85931779
    assert idx['New York, NY'][0].data.wof_id == 85977539





import os
import logging
import sys


LITECODER_ENV = os.environ.get('LITECODER_ENV')

DATA_DIR = os.path.join(os.path.dirname(__file__), 'data')

US_STATE_PATH = os.path.join(DATA_DIR, 'us-states.marisa')

US_CITY_PATH = os.path.join(DATA_DIR, 'us-cities.marisa')


logging.basicConfig(
    format='%(asctime)s | %(levelname)s : %(message)s',
    stream=sys.stdout,
    level=logging.INFO,
)

logger = logging.getLogger('litecoder')





import os


FIXTURES_DIR = os.path.join(os.path.dirname(__file__), 'fixtures')
REGION_DIR = os.path.join(FIXTURES_DIR, 'wof-region')
COUNTY_DIR = os.path.join(FIXTURES_DIR, 'wof-county')
LOCALITY_DIR = os.path.join(FIXTURES_DIR, 'wof-locality')





import os

from sqlalchemy.engine.url import URL
from sqlalchemy.orm import sessionmaker, scoped_session
from sqlalchemy import create_engine, event

from . import LITECODER_ENV, DATA_DIR


def connect_db(db_path):
    """Get database connection.

    Args:
        db_path (str)

    Returns: engine, session
    """
    url = URL(drivername='sqlite', database=db_path)
    engine = create_engine(url)

    # Fix transaction bugs in pysqlite.
    # http://docs.sqlalchemy.org/en/latest/dialects/sqlite.html#pysqlite-serializable

    @event.listens_for(engine, 'connect')
    def on_connect(conn, record):
        conn.execute('pragma foreign_keys=ON')
        conn.isolation_level = None

    @event.listens_for(engine, 'begin')
    def on_begin(conn):
        conn.execute('BEGIN')

    factory = sessionmaker(bind=engine)
    session = scoped_session(factory)

    return engine, session


db_path = os.path.join(DATA_DIR, 'litecoder.db')

# In-memory for tests.
if LITECODER_ENV == 'test':
    db_path = None

engine, session = connect_db(db_path)





from sqlalchemy import Column, Integer, String, Float

from .base import BaseModel


class WOFCounty(BaseModel):

    __tablename__ = 'wof_county'

    wof_id = Column(Integer, primary_key=True)

    wof_continent_id = Column(Integer)

    wof_country_id = Column(Integer)

    wof_region_id = Column(Integer)

    fips_code = Column(String)

    hasc_id = Column(String)

    wd_id = Column(String)

    name = Column(String)

    country_iso = Column(String)

    name_a0 = Column(String)

    name_a1 = Column(String)

    latitude = Column(Float)

    longitude = Column(Float)

    population = Column(Integer)

    area_m2 = Column(Float)





import attr
import os
import ujson
import time
import us

from collections import UserDict
from glob import iglob
from boltons.iterutils import chunked_iter
from multiprocessing import Pool
from tqdm import tqdm
from itertools import islice

from .. import logger, DATA_DIR
from ..utils import safe_property, first, read_json
from ..db import session
from ..models import WOFLocality, WOFRegion, WOFCounty


@attr.s
class WOFRepo:

    root = attr.ib()

    def paths_iter(self):
        """Glob paths.
        """
        pattern = os.path.join(self.root, '**/*.geojson')
        return iglob(pattern, recursive=True)

    def docs_iter(self, num_procs=None):
        """Generate parsed GeoJSON docs.
        """
        with Pool(num_procs) as p:
            yield from p.imap_unordered(read_json, self.paths_iter())

    def db_rows_iter(self):
        raise NotImplementedError

    def load_db(self, n=1000):
        """Load database rows.
        """
        rows = tqdm(self.db_rows_iter())

        for rows in chunked_iter(iter(rows), n):
            session.bulk_save_objects(rows)
            session.commit()


class WOFRegionRepo(WOFRepo):

    @classmethod
    def from_env(cls):
        return cls(os.path.join(DATA_DIR, 'wof-region'))

    def db_rows_iter(self):
        for doc in self.docs_iter():
            yield WOFRegionDoc(doc).db_row()


class WOFCountyRepo(WOFRepo):

    @classmethod
    def from_env(cls):
        return cls(os.path.join(DATA_DIR, 'wof-county'))

    def db_rows_iter(self):
        for doc in self.docs_iter():
            yield WOFCountyDoc(doc).db_row()


class WOFLocalityRepo(WOFRepo):

    @classmethod
    def from_env(cls):
        return cls(os.path.join(DATA_DIR, 'wof-locality'))

    def db_rows_iter(self):
        for doc in self.docs_iter():
            yield WOFLocalityDoc(doc).db_row()


class WOFDoc(UserDict):

    @classmethod
    def from_path(cls, path):
        return cls(read_json(path))

    def __repr__(self):
        return '%s<%d>' % (self.__class__.__name__, self.wof_id)

    @safe_property
    def wof_id(self):
        return self['id']

    def _wof_hierarchy_id(self, key):
        """Get WOF hierarchy id, if unique.
        """
        ids = set([
            h['%s_id' % key]
            for h in self['properties']['wof:hierarchy']
        ])

        # TODO: Handle 2+ links.
        # Punt when more than one link.
        if len(ids) == 1:
            id = list(ids)[0]
            return id if id > 0 else None

    @safe_property
    def wof_continent_id(self):
        return self._wof_hierarchy_id('continent')

    @safe_property
    def wof_country_id(self):
        return self._wof_hierarchy_id('country')

    @safe_property
    def wof_region_id(self):
        return self._wof_hierarchy_id('region')

    @safe_property
    def wof_county_id(self):
        return self._wof_hierarchy_id('county')

    @safe_property
    def dbp_id(self):
        return self['properties']['wof:concordances']['dbp:id']

    @safe_property
    def fb_id(self):
        return self['properties']['wof:concordances']['fb:id']

    @safe_property
    def fct_id(self):
        return self['properties']['wof:concordances']['fct:id']

    @safe_property
    def fips_code(self):
        return self['properties']['wof:concordances']['fips:code']

    @safe_property
    def gn_id(self):
        return self['properties']['wof:concordances']['gn:id']

    @safe_property
    def gp_id(self):
        return self['properties']['wof:concordances']['gp:id']

    @safe_property
    def hasc_id(self):
        return self['properties']['wof:concordances']['hasc:id']

    @safe_property
    def iso_id(self):
        return self['properties']['wof:concordances']['iso:id']

    @safe_property
    def unlc_id(self):
        return self['properties']['wof:concordances']['unlc:id']

    @safe_property
    def loc_id(self):
        return self['properties']['wof:concordances']['loc:id']

    @safe_property
    def nyt_id(self):
        return self['properties']['wof:concordances']['nyt:id']

    @safe_property
    def qs_id(self):
        return self['properties']['wof:concordances']['qs:id']

    @safe_property
    def qs_pg_id(self):
        return self['properties']['wof:concordances']['qs_pg:id']

    @safe_property
    def wd_id(self):
        return self['properties']['wof:concordances']['wd:id']

    @safe_property
    def wk_page(self):
        return self['properties']['wof:concordances']['wk:page']

    @safe_property
    def country_iso(self):
        return self['properties']['iso:country']

    @safe_property
    def _gn_latitude(self):
        return self['properties']['gn:latitude']

    @safe_property
    def _geom_latitude(self):
        return self['properties']['geom:latitude']

    @safe_property
    def latitude(self):
        return first(
            self._gn_latitude,
            self._geom_latitude,
        )

    @safe_property
    def _gn_longitude(self):
        return self['properties']['gn:longitude']

    @safe_property
    def _geom_longitude(self):
        return self['properties']['geom:longitude']

    @safe_property
    def longitude(self):
        return first(
            self._gn_longitude,
            self._geom_longitude,
        )

    @safe_property
    def _gn_population(self):
        return self['properties']['gn:population']

    @safe_property
    def _wof_population(self):
        return self['properties']['wof:population']

    @safe_property
    def _wk_population(self):
        return self['properties']['wk:population']

    @safe_property
    def population(self):
        return first(
            self._gn_population,
            self._wof_population,
            self._wk_population,
        )

    @safe_property
    def area_m2(self):
        return self['properties']['geom:area_square_m']

    @safe_property
    def _name_eng_x_preferred(self):
        return self['properties']['name:eng_x_preferred'][0]

    @safe_property
    def _wof_name(self):
        return self['properties']['wof:name']

    @safe_property
    def _qs_pg_name(self):
        return self['properties']['qs_pg:name']

    @safe_property
    def name(self):
        return first(
            self._name_eng_x_preferred,
            self._wof_name,
            self._qs_pg_name,
        )

    @safe_property
    def _qs_a0(self):
        return self['properties']['qs:a0']

    @safe_property
    def _qs_adm0(self):
        return self['properties']['qs:adm0']

    @safe_property
    def _ne_sov0name(self):
        return self['properties']['ne:SOV0NAME']

    @safe_property
    def _qs_pg_name_adm0(self):
        return self['properties']['qs_pg:name_adm0']

    @safe_property
    def _woe_name_adm0(self):
        return self['properties']['woe:name_adm0']

    @safe_property
    def name_a0(self):
        return first(
            self._qs_a0,
            self._qs_adm0,
            self._ne_sov0name,
            self._qs_pg_name_adm0,
            self._woe_name_adm0,
        )

    @safe_property
    def _qs_a1(self):
        return self['properties']['qs:a1'].strip('*')

    @safe_property
    def _ne_adm1name(self):
        return self['properties']['ne:ADM1NAME']

    @safe_property
    def _qs_pg_name_adm1(self):
        return self['properties']['qs_pg:name_adm1']

    @safe_property
    def _woe_name_adm1(self):
        return self['properties']['woe:name_adm1']

    @safe_property
    def name_a1(self):
        return first(
            self._qs_a1,
            self._ne_adm1name,
            self._qs_pg_name_adm1,
            self._woe_name_adm1,
        )


class WOFRegionDoc(WOFDoc):

    @safe_property
    def _abrv_eng_x_preferred(self):
        return self['properties']['abrv:eng_x_preferred'][0]

    @safe_property
    def _wof_abbreviation(self):
        return self['properties']['wof:abbreviation']

    @safe_property
    def name_abbr(self):
        return first(
            self._abrv_eng_x_preferred,
            self._wof_abbreviation,
        )

    def db_row(self):
        """Returns: models.WOFRegion
        """
        return WOFRegion(**{
            col: getattr(self, col)
            for col in WOFRegion.column_names()
            if hasattr(self, col)
        })


class WOFCountyDoc(WOFDoc):

    def db_row(self):
        """Returns: models.WOFCounty
        """
        return WOFCounty(**{
            col: getattr(self, col)
            for col in WOFCounty.column_names()
            if hasattr(self, col)
        })


class WOFLocalityDoc(WOFDoc):

    @safe_property
    def wikipedia_wordcount(self):
        return self['properties']['wk:wordcount']

    @safe_property
    def _gn_elevation(self):
        return self['properties']['gn:elevation']

    @safe_property
    def _ne_elevation(self):
        return self['properties']['ne:ELEVATION']

    @safe_property
    def elevation(self):
        return first(
            self._gn_elevation,
            self._ne_elevation,
        )

    def db_row(self):
        """Returns: models.WOFLocality
        """
        return WOFLocality(**{
            col: getattr(self, col)
            for col in WOFLocality.column_names()
            if hasattr(self, col)
        })





import ujson


class safe_property:

    @classmethod
    def cached(cls, func):
        return cls(func, True)

    def __init__(self, func, cached=False):
        self.cached = cached
        self.__doc__ = getattr(func, "__doc__")
        self.func = func

    def __get__(self, obj, cls):
        """Try to set a cached property. Catch and record errors.
        """
        if obj is None:
            return self

        try:
            value = self.func(obj)
        except Exception as e:
            value = None

        # Replace attribute with computed value.
        if self.cached:
            obj.__dict__[self.func.__name__] = value

        return value


def first(*seq):
    return next((x for x in seq if x is not None), None)


def read_json(path):
    with open(path) as fh:
        return ujson.load(fh)





import re
import marisa_trie
import ujson as json

from tqdm import tqdm
from collections import defaultdict
from itertools import product
from cached_property import cached_property
from box import Box

from sqlalchemy.inspection import inspect

from . import logger, US_CITY_PATH, US_STATE_PATH
from .models import WOFRegion, WOFLocality


# TODO: Country alt-names YAML.
USA_NAMES = (
    'USA',
    'United States',
    'United States of America',
    'US',
    'America',
)


def keyify(text):
    """Convert text -> normalized index key.
    """
    text = text.lower()
    text = text.strip()

    text = text.replace('.', '')
    text = re.sub('[,-]', ' ', text)
    text = re.sub('\s{2,}', ' ', text)

    return text


class CityNamePopulations(defaultdict):

    def __init__(self):
        """Index name -> [pops], using median pop if no metadata.
        """
        super().__init__(list)

        logger.info('Indexing name -> populations.')

        median_pop = WOFLocality.median_population()

        for row in tqdm(WOFLocality.query):
            for name in row.names:
                self[keyify(name)].append(row.population or median_pop)

    def __getitem__(self, text):
        return super().__getitem__(keyify(text))


class AllowBareCityName:

    def __init__(self, min_p1_gap=200000, blocklist=None):
        self.name_pops = CityNamePopulations()
        self.min_p1_gap = min_p1_gap
        self.blocklist = set(map(keyify, blocklist or []))

    def blocked(self, name):
        return keyify(name) in self.blocklist

    def large_p1_gap(self, row, name):
        """Get the difference in population between this city and the second-
        most-populous city with the name. Allow if over threshold.
        """
        all_pops = sorted(self.name_pops[name], reverse=True)
        pop = row.population or 0
        return pop - sum(all_pops[1:]) > self.min_p1_gap

    def __call__(self, row, name):
        """Is a name unique enough that it should be indexed independently?
        Args:
            row (models.WOFLocality)
            name (str)
        Returns: bool
        """
        return not self.blocked(name) and self.large_p1_gap(row, name)


class CityKeyIter:

    def __init__(self, allow_bare):
        self.allow_bare = allow_bare

    def __call__(self, row):
        """Enumerate index keys for a city.
        Args:
            row (db.Locality)
        Yields: str
        """
        bare_names = [n for n in row.names if self.allow_bare(row, n)]

        # Get non-empty state names.
        state_names = [n for n in (row.name_a1, row.us_state_abbr) if n]

        # Bare name
        for name in bare_names:
            yield name

        # Bare name, USA
        for name, usa in product(bare_names, USA_NAMES):
            yield ' '.join((name, usa))

        # Name, state
        for name, state in product(row.names, state_names):
            yield ' '.join((name, state))

        # Name, state, USA
        for name, state, usa in product(row.names, state_names, USA_NAMES):
            yield ' '.join((name, state, usa))


def state_key_iter(row):
    """Enumerate index keys for a state.
    Args:
        row (db.Region)
    Yields: str
    """
    names = (row.name,)
    abbrs = (row.name_abbr,)

    # Name
    yield from names

    # TODO: Bare abbrs?

    # Name, USA
    for name, usa in product(names, USA_NAMES):
        yield ' '.join((name, usa))

    # Abbr, USA
    for abbr, usa in product(abbrs, USA_NAMES):
        yield ' '.join((abbr, usa))


class Index:

    def load(self, path, mmap=False):
        if mmap:
            self._trie.mmap(path)
        else:
            self._trie.load(path)

    def __init__(self):
        self._trie = marisa_trie.BytesTrie()

        # We use prefixes here to store the keys -> ids and ids -> loc "maps" as subtrees in one marisa trie.
        self._keys_prefix = "A"
        self._ids_prefix = "B"

    def __len__(self):
        return len(self._trie.keys(self._keys_prefix))

    def __repr__(self):
        return '%s<%d keys, %d entities>' % (
            self.__class__.__name__,
            len(self._trie.keys(self._keys_prefix)),
            len(self._trie.keys(self._ids_prefix)),
        )

    def __getitem__(self, text):
        """Get ids, map to records only if there is a match in the index
        """
        normalized_key = self._keys_prefix + keyify(text)
        val = self._trie.get(normalized_key, None)
        if not val:
            return None
        ids = json.loads(val[0])

        return [json.loads(self._trie[self._ids_prefix + id][0]) for id in ids]

    def locations(self):
        return self._trie.items(self._ids_prefix)

    def save(self, path):
        self._trie.save(path)


class USCityIndex(Index):

    def load(self, path=US_CITY_PATH, mmap=False):
        return super().load(path, mmap)

    def __init__(self, bare_name_blocklist=None):
        super().__init__()
        self.bare_name_blocklist = bare_name_blocklist

    def build(self):
        """Index all US cities.
        """
        
        allow_bare = AllowBareCityName(blocklist=self.bare_name_blocklist)

        iter_keys = CityKeyIter(allow_bare)

        # Deduped cities.
        cities = WOFLocality.clean_us_cities()

        logger.info('Indexing US cities.')

        key_to_ids = defaultdict(set)
        id_to_loc_items = list()

        for row in tqdm(cities):

            # Key -> id(s)
            for key in map(keyify, iter_keys(row)):
                key_to_ids[key].add(str(row.wof_id))

            # ID -> city
            id_to_loc_items.append((self._ids_prefix + str(row.wof_id), bytes(json.dumps(dict(row)), encoding="utf-8")))

        key_to_ids_items = [(self._keys_prefix + key, json.dumps(list(key_to_ids[key])).encode("utf-8")) for key in key_to_ids]
        
        self._trie = marisa_trie.BytesTrie(key_to_ids_items + id_to_loc_items)


class USStateIndex(Index):

    def load(self, path=US_STATE_PATH, mmap=False):
        return super().load(path, mmap)

    def build(self):
        """Index all US states.
        """
        states = WOFRegion.query.filter(WOFRegion.country_iso=='US')

        logger.info('Indexing US states.')

        key_to_ids = defaultdict(set)
        id_to_loc_items = list()

        for row in tqdm(states):

            # Key -> id(s)
            for key in map(keyify, state_key_iter(row)):
                key_to_ids[key].add(str(row.wof_id))

            # ID -> state
            id_to_loc_items.append((self._ids_prefix + str(row.wof_id), bytes(json.dumps(dict(row)), encoding="utf-8")))

        key_to_ids_items = [(self._keys_prefix + key, json.dumps(list(key_to_ids[key])).encode("utf-8")) for key in key_to_ids]
        
        self._trie = marisa_trie.BytesTrie(key_to_ids_items + id_to_loc_items)





import yaml
import pkgutil
import us

import numpy as np

from tqdm import tqdm
from collections import defaultdict
from scipy.spatial import cKDTree

from sqlalchemy import Column, Integer, String, Float, Boolean
from sqlalchemy.orm import relationship

from ..db import session
from ..utils import safe_property
from .. import logger
from .base import BaseModel


# TODO: Make pluggable.
CITY_ALT_NAMES = yaml.load(pkgutil.get_data(
    'litecoder', 'data/city-alt-names.yml'
))


ID_COLS = (
    'dbp_id',
    'fb_id',
    'fct_id',
    'fips_code',
    'gn_id',
    'gp_id',
    'loc_id',
    'nyt_id',
    'qs_id',
    'qs_pg_id',
    'wd_id',
    'wk_page',
)


class WOFLocality(BaseModel):

    __tablename__ = 'wof_locality'

    wof_id = Column(Integer, primary_key=True)

    wof_continent_id = Column(Integer)

    wof_country_id = Column(Integer)

    wof_region_id = Column(Integer)

    wof_county_id = Column(Integer)

    dbp_id = Column(String)

    fb_id = Column(String)

    fct_id = Column(String)

    fips_code = Column(String)

    gn_id = Column(Integer)

    gp_id = Column(Integer)

    loc_id = Column(String)

    nyt_id = Column(String)

    qs_id = Column(Integer)

    qs_pg_id = Column(Integer)

    wd_id = Column(String)

    wk_page = Column(String)

    name = Column(String)

    country_iso = Column(String)

    name_a0 = Column(String)

    name_a1 = Column(String)

    latitude = Column(Float)

    longitude = Column(Float)

    population = Column(Integer)

    wikipedia_wordcount = Column(Integer)

    elevation = Column(Integer)

    area_m2 = Column(Float)

    duplicate = Column(Boolean, default=False)

    # TODO: Handle NYC, with separate county for each borough.
    county = relationship(
        'WOFCounty',
        primaryjoin='WOFCounty.wof_id==WOFLocality.wof_county_id',
        foreign_keys=[wof_county_id],
    )

    region = relationship(
        'WOFRegion',
        primaryjoin='WOFRegion.wof_id==WOFLocality.wof_region_id',
        foreign_keys=[wof_region_id],
    )

    @classmethod
    def set_dupes(cls, wof_ids):
        """Apply duplicate list.

        Args:
            wof_ids (iter)
        """
        dupes = cls.query.filter(cls.wof_id.in_(wof_ids))
        dupes.update({cls.duplicate: True}, synchronize_session=False)
        session.commit()

    @classmethod
    def dedupe_id_col(cls, name):
        """Dedupe localities via identifier column.
        """
        dup_col = getattr(cls, name)

        logger.info('Mapping `%s` -> rows' % name)

        id_rows = defaultdict(list)
        for row in tqdm(cls.query.filter(dup_col != None)):
            id_rows[getattr(row, name)].append(row)

        logger.info('Deduping rows with shared `%s`' % name)

        dupes = set()
        for rows in tqdm(id_rows.values()):
            if len(rows) > 1:

                # Sort by completeness.
                rows = sorted(rows, key=lambda r: r.field_count, reverse=True)

                # Add all but most complete to dupes.
                for row in rows[1:]:
                    dupes.add(row.wof_id)

        cls.set_dupes(dupes)

    @classmethod
    def dedupe_proximity(cls, buffer=0.1):
        """Find duplicates within N degrees.
        """
        # Pre-load rows.
        rows = cls.query.all()
        id_row = {i: row for i, row in enumerate(rows)}

        # Build index.
        data = [[r.longitude, r.latitude] for r in rows]
        idx = cKDTree(data)

        logger.info('Deduping rows within %.2f°' % buffer)

        dupes = set()
        for id1, id2 in tqdm(idx.query_pairs(buffer)):

            row1, row2 = id_row[id1], id_row[id2]

            if row1.name == row2.name:

                dupes.add(row1.wof_id if row1.field_count < row2.field_count
                    else row2.wof_id)

        cls.set_dupes(dupes)

    @classmethod
    def dedupe(cls, *args, **kwargs):
        """Dedupe by shared ids + proximity.
        """
        for name in ID_COLS:
            cls.dedupe_id_col(name)

        cls.dedupe_proximity(*args, **kwargs)

    @classmethod
    def clean_query(cls):
        """Build deduped query.
        """
        return cls.query.filter(cls.duplicate==False)

    @classmethod
    def clean_us_cities(cls):
        """Select clean US cities with city + state names.
        """
        return (cls.clean_query()
            .filter(cls.country_iso=='US')
            .filter(cls.name != None)
            .filter(cls.name_a1 != None))

    @classmethod
    def median_population(cls):
        """Get median population.
        """
        pops = [c.population for c in cls.query if c.population]
        return np.median(pops)

    def __repr__(self):
        return '%s<%s, %s, %s, wof:%d>' % (
            self.__class__.__name__,
            self.name, self.name_a1, self.name_a0, self.wof_id,
        )

    @property
    def field_count(self):
        """Count non-null fields.
        """
        return len([k for k, v in dict(self).items() if v is not None])

    # TODO: Make name list pluggable.
    @property
    def alt_names(self):
        """Map alt names via Wikidata id.
        """
        return CITY_ALT_NAMES.get(self.wd_id, [])

    @property
    def names(self):
        """Name + alt names.
        """
        return set((self.name, *self.alt_names))

    # TODO: Get from region?
    @safe_property
    def us_state_abbr(self):
        """A1 name -> US state abbreviation.
        """
        return us.states.lookup(self.name_a1).abbr





import pytest

from litecoder.usa import keyify


@pytest.mark.parametrize('text,key', [

    # Downcase
    ('BOSTON', 'boston'),

    # Strip
    ('  Boston  ', 'boston'),

    # Remove periods
    ('Washington D.C.', 'washington dc'),

    # Comma -> space
    ('Boston,MA', 'boston ma'),

    # Dash -> space
    ('La-La Land', 'la la land'),

    # 2+ whitespace -> 1 space
    ('Boston  MA   USA', 'boston ma usa'),

])
def test_keyify(text, key):
    assert keyify(text) == key





import pytest

from litecoder.models import WOFLocality

from tests.utils import read_yaml


def yield_cases():
    """Generate cases from YAML file.
    """
    cases = read_yaml(__file__, 'test_us_city_index.yml')

    for group in cases:

        queries = group['query']

        xfail = group.get('xfail', False)

        if type(queries) is str:
            queries = [queries]

        for query in queries:
            yield query, group['matches'], xfail


@pytest.mark.parametrize('query,matches,xfail', yield_cases())
def test_cases(city_idx, query, matches, xfail):

    if xfail:
        pytest.xfail()

    res = city_idx[query]

    ids = [r["wof_id"] for r in res]

    # Exact id list match.
    assert sorted(ids) == sorted(matches)


topn = (WOFLocality.clean_us_cities()
    .order_by(WOFLocality.population.desc())
    .limit(1000))


@pytest.mark.parametrize('city', topn)
def test_topn(city_idx, city):
    """Smoke test N most populous cities.
    """
    res = city_idx['%s, %s' % (city.name, city.name_a1)]
    res_ids = [r["wof_id"] for r in res]

    assert city.wof_id in res_ids




import pytest
import os

from litecoder.models import WOFRegion

from tests.utils import read_yaml


pytestmark = pytest.mark.usefixtures('load_db')


cases = read_yaml(__file__, 'test_ingest_region.yml')


@pytest.mark.parametrize('wof_id,fields', cases.items())
def test_test(wof_id, fields):

    row = WOFRegion.query.get(wof_id)

    for col, val in fields.items():

        if type(val) is float:
            val = pytest.approx(val, 0.001)

        assert getattr(row, col) == val





import pytest
import os

os.environ['LITECODER_ENV'] = 'test'

from litecoder.db import engine, session
from litecoder.models import BaseModel
from litecoder.usa import USCityIndex, USStateIndex

from litecoder.sources.wof import (
    WOFRegionRepo, WOFCountyRepo, WOFLocalityRepo
)

from . import REGION_DIR, COUNTY_DIR, LOCALITY_DIR


@pytest.fixture(scope='module')
def reset_db():
    """Drop and recreate the tables.
    """
    BaseModel.metadata.drop_all(engine)
    BaseModel.metadata.create_all(engine)


@pytest.fixture(scope='module')
def load_db(reset_db):
    """Load tables.
    """
    WOFRegionRepo(REGION_DIR).load_db()
    WOFCountyRepo(COUNTY_DIR).load_db()
    WOFLocalityRepo(LOCALITY_DIR).load_db()



from setuptools import setup, find_packages


INSTALL_REQUIRES = [
    "marisa-trie>=0.7.5",
    "ujson",
    "tqdm",
    "cached_property",
    "sqlalchemy",
    "download",
]

CLASSIFIERS = [
    "Development Status :: 3 - Alpha",
    "Intended Audience :: Developers",
    "License :: OSI Approved :: MIT License",
    "Programming Language :: Python",
    "Programming Language :: Python :: 3.5",
    "Programming Language :: Python :: 3.6",
    "Programming Language :: Python :: 3.7",
]


setup(
    name="litecoder",
    version="0.3.0",
    description="US city + state geocoding.",
    url="https://github.com/social-machines/litecoder",
    download_url="https://github.com/social-machines/litecoder/archive/v0.3.0.tar.gz",
    license="MIT",
    author="Lab for Social Machines, MIT Media Lab",
    author_email="wesc@media.mit.edu",
    packages=find_packages(),
    classifiers=CLASSIFIERS,
    install_requires=INSTALL_REQUIRES,
)





import pytest

from litecoder.usa import USCityIndex, USStateIndex


@pytest.fixture(scope='session')
def city_idx():
    city_idx = USCityIndex()
    city_idx.load()
    return city_idx


@pytest.fixture(scope='session')
def state_idx():
    state_idx = USStateIndex()
    state_idx.load()
    return state_idx
